@@ -48,7 +48,7 @@ int Detail(char *dev, int brief, int export, int test, char *homehost)
 	int is_26 = get_linux_version() >= 2006000;
 	int is_rebuilding = 0;
 	int failed = 0;
-	struct supertype *st = NULL;
+	struct supertype *st;
 	int max_disks = MD_SB_DISKS; /* just a default */
 	struct mdinfo info;
 	struct mdinfo *sra;
@@ -111,9 +111,11 @@ int Detail(char *dev, int brief, int export, int test, char *homehost)
 	}
 
 	/* try to load a superblock */
-	for (d= 0; d<max_disks; d++) {
+	if (st) for (d = 0; d < max_disks; d++) {
 		mdu_disk_info_t disk;
 		char *dv;
+		int fd2;
+		int err;
 		disk.number = d;
 		if (ioctl(fd, GET_DISK_INFO, &disk) < 0)
 			continue;
@@ -121,38 +123,49 @@ int Detail(char *dev, int brief, int export, int test, char *homehost)
 		    disk.major == 0 &&
 		    disk.minor == 0)
 			continue;
-		if ((dv=map_dev(disk.major, disk.minor, 1))) {
-			/* some formats (imsm) have free-floating-spares
-			 * with a uuid of uuid_match_any, they don't
-			 * have very good info about the rest of the
-			 * container, so keep searching when
-			 * encountering such a device.  Otherwise, stop
-			 * after the first successful call to
-			 * ->load_super.
-			 */
-			int free_spare = memcmp(uuid_match_any,
-						info.uuid,
-						sizeof(uuid_match_any)) == 0;
-			if ((!st || !st->sb || free_spare) &&
-			    (array.raid_disks == 0 || 
-			     (disk.state & (1<<MD_DISK_ACTIVE)))) {
-				/* try to read the superblock from this device
-				 * to get more info
-				 */
-				if (free_spare)
-					st->ss->free_super(st);
-				int fd2 = dev_open(dv, O_RDONLY);
-				if (fd2 >=0 && st &&
-				    st->ss->load_super(st, fd2, NULL) == 0) {
-					st->ss->getinfo_super(st, &info);
-					if (array.raid_disks != 0 && /* container */
-					    (info.array.ctime != array.ctime ||
-					     info.array.level != array.level))
-						st->ss->free_super(st);
-				}
-				if (fd2 >= 0) close(fd2);
-			}
+
+		if (array.raid_disks > 0 &&
+		    (disk.state & (1 << MD_DISK_ACTIVE)) == 0)
+			continue;
+
+		dv = map_dev(disk.major, disk.minor, 1);
+		if (!dv)
+			continue;
+
+		fd2 = dev_open(dv, O_RDONLY);
+		if (fd2 < 0)
+			continue;
+
+		if (st->sb)
+			st->ss->free_super(st);
+
+		err = st->ss->load_super(st, fd2, NULL);
+		close(fd2);
+		if (err)
+			continue;
+		st->ss->getinfo_super(st, &info);
+
+		if (array.raid_disks != 0 && /* container */
+		    (info.array.ctime != array.ctime ||
+		     info.array.level != array.level)) {
+			st->ss->free_super(st);
+			continue;
 		}
+		/* some formats (imsm) have free-floating-spares
+		 * with a uuid of uuid_match_any, they don't
+		 * have very good info about the rest of the
+		 * container, so keep searching when
+		 * encountering such a device.  Otherwise, stop
+		 * after the first successful call to
+		 * ->load_super.
+		 */
+		if (memcmp(uuid_match_any,
+			   info.uuid,
+			   sizeof(uuid_match_any)) == 0) {
+			st->ss->free_super(st);
+			continue;
+		}
+		break;
 	}
 
 	/* Ok, we have some info to print... */
@@ -174,15 +187,14 @@ int Detail(char *dev, int brief, int export, int test, char *homehost)
 			if (sra && sra->array.major_version < 0)
 				printf("MD_METADATA=%s\n", sra->text_version);
 			else
-				printf("MD_METADATA=%d.%02d\n",
+				printf("MD_METADATA=%d.%d\n",
 				       array.major_version, array.minor_version);
 		}
 		
 		if (st && st->sb) {
-			struct mdinfo info;
 			char nbuf[64];
 			struct map_ent *mp, *map = NULL;
-			st->ss->getinfo_super(st, &info);
+
 			fname_from_uuid(st, &info, nbuf, ':');
 			printf("MD_UUID=%s\n", nbuf+5);
 			mp = map_by_uuid(&map, info.uuid);
@@ -207,6 +219,42 @@ int Detail(char *dev, int brief, int export, int test, char *homehost)
 		goto out;
 	}
 
+	disks = malloc(max_disks * sizeof(mdu_disk_info_t));
+	for (d=0; d<max_disks; d++) {
+		disks[d].state = (1<<MD_DISK_REMOVED);
+		disks[d].major = disks[d].minor = 0;
+		disks[d].number = disks[d].raid_disk = d;
+	}
+
+	next = array.raid_disks;
+	for (d=0; d < max_disks; d++) {
+		mdu_disk_info_t disk;
+		disk.number = d;
+		if (ioctl(fd, GET_DISK_INFO, &disk) < 0) {
+			if (d < array.raid_disks)
+				fprintf(stderr, Name ": cannot get device detail for device %d: %s\n",
+					d, strerror(errno));
+			continue;
+		}
+		if (disk.major == 0 && disk.minor == 0)
+			continue;
+		if (disk.raid_disk >= 0 && disk.raid_disk < array.raid_disks)
+			disks[disk.raid_disk] = disk;
+		else if (next < max_disks)
+			disks[next++] = disk;
+	}
+
+	avail = calloc(array.raid_disks, 1);
+
+	for (d= 0; d < array.raid_disks; d++) {
+		mdu_disk_info_t disk = disks[d];
+
+		if ((disk.state & (1<<MD_DISK_SYNC))) {
+			avail_disks ++;
+			avail[d] = 1;
+		}
+	}
+
 	if (brief) {
 		mdu_bitmap_file_t bmf;
 		printf("ARRAY %s", dev);
@@ -226,7 +274,7 @@ int Detail(char *dev, int brief, int export, int test, char *homehost)
 			if (sra && sra->array.major_version < 0)
 				printf(" metadata=%s", sra->text_version);
 			else
-				printf(" metadata=%d.%02d",
+				printf(" metadata=%d.%d",
 				       array.major_version, array.minor_version);
 		}
 
@@ -242,7 +290,7 @@ int Detail(char *dev, int brief, int export, int test, char *homehost)
 		struct mdstat_ent *ms = mdstat_read(0, 0);
 		struct mdstat_ent *e;
 		int devnum = array.md_minor;
-		if (major(stb.st_rdev) == get_mdp_major())
+		if (major(stb.st_rdev) == (unsigned)get_mdp_major())
 			devnum = -1 - devnum;
 
 		for (e=ms; e; e=e->next)
@@ -259,7 +307,7 @@ int Detail(char *dev, int brief, int export, int test, char *homehost)
 		if (sra && sra->array.major_version < 0)
 			printf("        Version : %s\n", sra->text_version);
 		else
-			printf("        Version : %d.%02d\n",
+			printf("        Version : %d.%d\n",
 			       array.major_version, array.minor_version);
 		}
 
@@ -306,13 +354,23 @@ int Detail(char *dev, int brief, int export, int test, char *homehost)
 		atime = array.utime;
 		if (atime)
 			printf("    Update Time : %.24s\n", ctime(&atime));
-		if (array.raid_disks)
+		if (array.raid_disks) {
+			char *st;
+			if (avail_disks == array.raid_disks)
+				st = "";
+			else if (!enough(array.level, array.raid_disks,
+					 array.layout, 1, avail, avail_disks))
+				st = ", FAILED";
+			else
+				st = ", degraded";
+
 			printf("          State : %s%s%s%s\n",
 			       (array.state&(1<<MD_SB_CLEAN))?"clean":"active",
-			       array.active_disks < array.raid_disks? ", degraded":"",
+			       st,
 			       (!e || e->percent < 0) ? "" :
 			       (e->resync) ? ", resyncing": ", recovering",
 			       larray_size ? "": ", Not Started");
+		}
 		if (array.raid_disks)
 			printf(" Active Devices : %d\n", array.active_disks);
 		printf("Working Devices : %d\n", array.working_disks);
@@ -363,7 +421,7 @@ int Detail(char *dev, int brief, int export, int test, char *homehost)
 #if 0
 This is pretty boring
 			printf("  Reshape pos'n : %llu%s\n", (unsigned long long) info.reshape_progress<<9,
-			       human_size(info.reshape_progress<<9));
+			       human_size((unsigned long long)info.reshape_progress<<9));
 #endif
 			if (info.delta_disks > 0)
 				printf("  Delta Devices : %d, (%d->%d)\n",
@@ -382,6 +440,11 @@ This is pretty boring
 					printf("     New Layout : %s\n",
 					       c?c:"-unknown-");
 				}
+				if (info.new_level == 6) {
+					char *c = map_num(r6layout, info.new_layout);
+					printf("     New Layout : %s\n",
+					       c?c:"-unknown-");
+				}
 				if (info.new_level == 10) {
 					printf("     New Layout : near=%d, %s=%d\n",
 					       info.new_layout&255,
@@ -437,32 +500,7 @@ This is pretty boring
 		else
 			printf("    Number   Major   Minor   RaidDevice\n");
 	}
-	disks = malloc(max_disks * sizeof(mdu_disk_info_t));
-	for (d=0; d<max_disks; d++) {
-		disks[d].state = (1<<MD_DISK_REMOVED);
-		disks[d].major = disks[d].minor = 0;
-		disks[d].number = disks[d].raid_disk = d;
-	}
 
-	next = array.raid_disks;
-	for (d=0; d < max_disks; d++) {
-		mdu_disk_info_t disk;
-		disk.number = d;
-		if (ioctl(fd, GET_DISK_INFO, &disk) < 0) {
-			if (d < array.raid_disks)
-				fprintf(stderr, Name ": cannot get device detail for device %d: %s\n",
-					d, strerror(errno));
-			continue;
-		}
-		if (disk.major == 0 && disk.minor == 0)
-			continue;
-		if (disk.raid_disk >= 0 && disk.raid_disk < array.raid_disks)
-			disks[disk.raid_disk] = disk;
-		else if (next < max_disks)
-			disks[next++] = disk;
-	}
-
-	avail = calloc(array.raid_disks, 1);
 	for (d= 0; d < max_disks; d++) {
 		char *dv;
 		mdu_disk_info_t disk = disks[d];
@@ -515,11 +553,6 @@ This is pretty boring
 		if (test && d < array.raid_disks
 		    && !(disk.state & (1<<MD_DISK_SYNC)))
 			rv |= 1;
-		if (d < array.raid_disks
-		    && (disk.state & (1<<MD_DISK_SYNC))) {
-			avail_disks ++;
-			avail[d] = 1;
-		}
 		if ((dv=map_dev(disk.major, disk.minor, 0))) {
 			if (brief) {
 				if (devices) {
